@@ -36,6 +36,7 @@ import fsl.utils.tempdir                  as tempdir
 import fsl.utils.idle                     as idle
 
 import                                       fsleyes
+import fsleyes.main                       as fsleyes_main
 import fsleyes.strings                    as strings
 import fsleyes.actions.screenshot         as screenshot
 
@@ -46,7 +47,6 @@ try:
     import                            zmq
 
     import zmq.eventloop.zmqstream as zmqstream
-    import tornado.ioloop          as ioloop
 
     import ipykernel.ipkernel      as ipkernel
     import ipykernel.iostream      as iostream
@@ -95,24 +95,34 @@ class NotebookAction(base.Action):
         :arg displayCtx:  The master :class:`.DisplayContext`.
         :arg frame:       The :class:`.FSLeyesFrame`.
         """
-        base.Action.__init__(self, self.__openNotebooks)
+        base.Action.__init__(
+            self, overlayList, displayCtx, self.__openNotebooks)
 
         # permanently disable if any
         # dependencies are not present
-        self.enabled        = ENABLED
-        self.__overlayList  = overlayList
-        self.__displayCtx   = displayCtx
-        self.__frame        = frame
-        self.__kernel       = None
-        self.__server       = None
+        self.enabled  = ENABLED
+        self.__frame  = frame
+        self.__kernel = None
+        self.__server = None
 
 
-    def __openNotebooks(self):
+    def __openNotebooks(self, nbfile=None):
         """Called when this ``NotebookAction`` is invoked. Starts the
-        server and kernel if necessary, then opens a new notebook in a
-        web browser.
+        server and kernel if necessary.
+
+        If the server/kernel have not yet been started and ``nbfile`` is
+        provided, the server will be started with ``nbfile`` opened.
         """
 
+        # If the kernel and server are both
+        # ok, open the notebook homepage
+        if self.__kernel is not None and \
+           self.__server is not None and \
+           self.__kernel.is_alive()  and \
+           self.__server.is_alive():
+            webbrowser.open(self.__server.url)
+            return
+
         # have the kernel or server threads crashed?
         if self.__kernel is not None and not self.__kernel.is_alive():
             self.__kernel = None
@@ -136,27 +146,22 @@ class NotebookAction(base.Action):
                 if self.__kernel is None:
                     self.__kernel = self.__startKernel(progdlg)
                 if self.__server is None:
-                    self.__server = self.__startServer(progdlg)
+                    self.__server = self.__startServer(progdlg, nbfile)
 
         finally:
             if progdlg is not None:
                 progdlg.Destroy()
                 progdlg = None
 
-        # if all is well, open the
-        # notebook server homepage
-        webbrowser.open(self.__server.url)
-
 
     def __bounce(self, secs, progdlg):
         """Used by :meth:`__startKernel` and :meth:`__startServer`. Blocks for
-        from ipykernel.kernelapp import IPKernelApp``secs``, bouncing
-        the :class:`.Progress` dialog ten times per second, and yielding to
-        the ``wx`` main loop.
+        ``secs``, bouncing the :class:`.Progress` dialog ten times per second,
+        and yielding to the ``wx`` main loop.
         """
         for i in range(int(secs * 10)):
             progdlg.DoBounce()
-            wx.Yield()
+            wx.GetApp().Yield()
             time.sleep(0.1)
 
 
@@ -169,11 +174,11 @@ class NotebookAction(base.Action):
         """
         progdlg.UpdateMessage(strings.messages[self, 'init.kernel'])
         kernel = BackgroundIPythonKernel(
-            self.__overlayList,
-            self.__displayCtx,
+            self.overlayList,
+            self.displayCtx,
             self.__frame)
         kernel.start()
-        self.__bounce(1, progdlg)
+        self.__bounce(2, progdlg)
 
         if not kernel.is_alive():
             raise RuntimeError('Could not start IPython kernel: '
@@ -182,7 +187,7 @@ class NotebookAction(base.Action):
         return kernel
 
 
-    def __startServer(self, progdlg):
+    def __startServer(self, progdlg, nbfile=None):
         """Attempts to create and start a :class:`NotebookServer`.
 
         :returns: the server if it was started.
@@ -191,23 +196,37 @@ class NotebookAction(base.Action):
         """
 
         progdlg.UpdateMessage(strings.messages[self, 'init.server'])
-        server = NotebookServer(self.__kernel.connfile)
+        server = NotebookServer(self.__kernel.connfile, nbfile)
         server.start()
-        self.__bounce(1.5, progdlg)
 
-        if not server.is_alive():
+        elapsed = 0
+
+        while elapsed < 5 and not server.ready:
+            self.__bounce(0.5, progdlg)
+            elapsed += 0.5
+
+        if elapsed >= 5 or not server.is_alive():
             raise RuntimeError('Could not start notebook server: '
                                '{}'.format(server.stderr))
 
         return server
 
 
-class BackgroundIPythonKernel(threading.Thread):
+class BackgroundIPythonKernel(object):
     """The BackgroundIPythonKernel creates an IPython jupyter kernel and makes
-    it accessible over tcp (on the local machine only). The ``zmq`` event loop
-    is run on a separate thread, but the kernel handles events on the main
-    thread, via :func:`.idle.idle`.
+    it accessible over tcp (on the local machine only).
 
+    Before FSLeyes 0.28.0, this class derived from ``threading.Thread``, and
+    ran its own ``IOLoop`` which was used to run the IPython kernel.  But now
+    the IPython kernel is run on the main thread, by repeatedly scheduling
+    calls to ``IPythonKernel.do_one_iteration`` on :func:`.idle.idle`.
+
+    This update was necessary due to changes in ``ipykernel`` from version 4
+    to version 5, namely that ipykernel 5 uses ``asyncio`` co-routines, making
+    it difficult to transfer execution between threads.
+
+    Because the ``BackgroundIPythonKernel`` used to be a thread, it is still
+    necessary to call its :meth:`start` method to start the kernel loop.
 
     The Jupyter/IPython documentation is quite fragmented at this point in
     time; `this github issue <https://github.com/ipython/ipython/issues/8097>`_
@@ -226,16 +245,13 @@ class BackgroundIPythonKernel(threading.Thread):
         :arg frame:       The :class:`.FSLeyesFrame`.
         """
 
-        threading.Thread.__init__(self)
-
-        self.daemon        = True
         ip                 = '127.0.0.1'
         transport          = 'tcp'
         addr               = '{}://{}'.format(transport, ip)
         self.__connfile    = None
-        self.__ioloop      = None
         self.__kernel      = None
         self.__error       = None
+        self.__lastIter    = 0
         self.__overlayList = overlayList
         self.__displayCtx  = displayCtx
         self.__frame       = frame
@@ -318,6 +334,24 @@ class BackgroundIPythonKernel(threading.Thread):
         atexit.register(os.remove, self.__connfile)
 
 
+    def is_alive(self):
+        """Returns ``True`` if the kernel loop appears to be running, ``False``
+        otherwise
+        """
+        # We check that the last kernel
+        # iteration was not too long ago,
+        # that an error hasn't occurred,
+        # and that the worker threads are
+        # alive.
+        now   = time.time()
+        delta = 5.0
+
+        return (now - self.__lastIter) < delta and \
+               self.__error is None            and \
+               self.__iopub.thread.is_alive()  and \
+               self.__heartbeat.is_alive()
+
+
     @property
     def kernel(self):
         """The ``IPythonKernel`` object. """
@@ -340,8 +374,8 @@ class BackgroundIPythonKernel(threading.Thread):
 
     @property
     def error(self):
-        """If an error occurs on the background thread causing it to crash,
-        a reference to the ``Exception`` is stored here.
+        """If an error occurs on the kernel loop causing it to crash, a
+        reference to the ``Exception`` is stored here.
         """
         return self.__error
 
@@ -353,30 +387,36 @@ class BackgroundIPythonKernel(threading.Thread):
             return display.Image('screenshot.png')
 
 
-    def __kernelDispatch(self):
-        """Event loop used for the IPython kernel. Submits the kernel function
-        to the :func:`.idle.idle` loop, and schedules another call to this
-        method on the ``zmq`` event loop.
-
-        This means that, while the ``zmq`` loop runs in its own thread, the
-        IPython kernel is executed on the main thread.
+    def start(self):
+        """Start the IPython kernel loop. This method returns immediately - the
+        kernel loop is driven on :func:`.idle.idle`.
         """
-
-        idle.idle(self.__kernel.do_one_iteration)
-        self.__ioloop.call_later(self.__kernel._poll_interval,
-                                 self.__kernelDispatch)
+        self.__kernel.start()
+        idle.idle(self.__eventloop, after=self.__kernel._poll_interval)
 
 
-    def run(self):
-        """Start the IPython kernel and Run the ``zmq`` event loop. """
+    def __eventloop(self):
+        """Event loop used for the IPython kernel. Calls :meth:`__kernelDispatch`,
+        then schedules a future call to ``__eventloop`` via :func:`.idle.idle`
+        loop.
+        """
+        self.__kernelDispatch()
+        idle.idle(self.__eventloop, after=self.__kernel._poll_interval)
 
+
+    def __kernelDispatch(self):
+        """Execute one kernel iteration, by scheduling a call to
+        ``IPythonKernel.do_one_iteration`` on the kernel's io loop.
+        """
         try:
-            self.__ioloop = ioloop.IOLoop()
-            self.__ioloop.make_current()
-            self.__kernel.start()
-            self.__ioloop.call_later(self.__kernel._poll_interval,
-                                     self.__kernelDispatch)
-            self.__ioloop.start()
+            loop = self.__kernel.io_loop
+            loop.run_sync(self.__kernel.do_one_iteration)
+
+            # save the time on each iteration,
+            # so the is_alive method has an
+            # idea of whether the kernel is
+            # still running
+            self.__lastIter = time.time()
 
         except Exception as e:
             self.__error = e
@@ -451,15 +491,18 @@ class NotebookServer(threading.Thread):
     """
 
 
-    def __init__(self, connfile):
+    def __init__(self, connfile, nbfile=None):
         """Create a ``NotebookServer`` thread.
 
         :arg connfile: Connection file of the IPython kernel to connect to.
+        :arg nbfile:   Path to a notebook file which should be opened on
+                       startup.
         """
 
         threading.Thread.__init__(self)
         self.daemon        = True
         self.__connfile    = connfile
+        self.__nbfile      = nbfile
         self.__stdout      = None
         self.__stderr      = None
         self.__port        = None
@@ -474,12 +517,23 @@ class NotebookServer(threading.Thread):
         if self.__port is not None:
             return self.__port
 
+        self.__port = self.__readPort()
+        return self.__port
+
+
+    def __readPort(self):
         for server in notebookapp.list_running_servers():
             if server['token'] == self.__token:
-                self.__port = server['port']
-                break
+                return server['port']
+        return None
 
-        return self.__port
+
+    @property
+    def ready(self):
+        """Returns ``True`` if the server is running and ready, ``False``
+        otherwise.
+        """
+        return self.__readPort() is not None
 
 
     @property
@@ -534,52 +588,54 @@ class NotebookServer(threading.Thread):
         # JUPYTER_CONFIG_DIR is set, so our
         # custom bits and pieces will be found.
         env        = dict(os.environ)
-        pythonpath = os.pathsep.join((cfgdir, env['PYTHONPATH']))
+        pythonpath = os.pathsep.join((cfgdir, env.get('PYTHONPATH', '')))
 
         env['JUPYTER_CONFIG_DIR'] = cfgdir
         env['PYTHONPATH']         = pythonpath
 
-        # command to start the notebook
-        # server in a sub-process
-
-        # With frozen FSLeyes versions, there
-        # probbaly isn't a 'jupyter-notebook'
-        # executable. So we use a hook in
-        # fsleyes.main to run the server.
+        # command to start the notebook server
+        # in a sub-process - we run the server
+        # via a wrapper function called nbmain,
+        # defined below.
+        #
+        # But with frozen FSLeyes versions, we
+        # can't just call a python interpreter,
+        # so we use a hook in fsleyes.main to
+        # call actions.notebook_main.main.
         if fslplatform.frozen:
-            exe = op.join(op.dirname(sys.executable), 'fsleyes')
-            log.debug('Running notebook server via %s notebook', sys.argv[0])
-
-            # py2app manipulates the PYTHONPATH, so we
-            # pass it through as a command-line argument.
-            self.__nbproc = sp.Popen([exe, 'notebook', cfgdir],
-                                     stdout=sp.PIPE,
-                                     stderr=sp.PIPE,
-                                     cwd=cfgdir,
-                                     env=env)
-
-        # Otherwise we can call
-        # it in the usual manner.
+            cmd = [op.join(op.dirname(sys.executable), 'fsleyes')]
         else:
-            log.debug('Running notebook server via jupyter-notebook')
-            self.__nbproc = sp.Popen(['jupyter-notebook'],
-                                     stdout=sp.PIPE,
-                                     stderr=sp.PIPE,
-                                     cwd=cfgdir,
-                                     env=env)
+            cmd = [sys.executable, fsleyes_main.__file__]
+
+        # py2app manipulates the PYTHONPATH,
+        # so we pass the cfgdir through as a
+        # command-line argument - it is picked
+        # up again by the nbmain function.
+        cmd.extend(('notebook', 'server', cfgdir))
+
+        if self.__nbfile is not None:
+            cmd.append(self.__nbfile)
+
+        log.debug('Running notebook server via %s notebook', cmd[0])
+        self.__nbproc = sp.Popen(cmd,
+                                 stdout=sp.PIPE,
+                                 stderr=sp.PIPE,
+                                 env=env)
 
         def killServer():
-            # We need two CTRL+Cs to kill
-            # the notebook server
-            self.__nbproc.terminate()
-            self.__nbproc.terminate()
+            if self.__nbproc is not None:
+                # We need two CTRL+Cs to kill
+                # the notebook server
+                self.__nbproc.terminate()
+                self.__nbproc.terminate()
 
-            # Give it a little time, then
-            # force kill if needed
-            try:
-                self.__nbproc.wait(0.25)
-            except sp.TimeoutExpired:
-                self.__nbproc.kill()
+                # Give it a little time, then
+                # force kill if needed
+                try:
+                    self.__nbproc.wait(0.25)
+                except sp.TimeoutExpired:
+                    self.__nbproc.kill()
+                self.__nbproc = None
 
             shutil.rmtree(op.abspath(op.join(cfgdir, '..')))
 
@@ -590,6 +646,14 @@ class NotebookServer(threading.Thread):
         o, e          = self.__nbproc.communicate()
         self.__stdout = o.decode()
         self.__stderr = e.decode()
+        self.__nbproc = None
+
+        # if we've gotten this far,
+        # call our atexit handler
+        # directly, and unregister
+        # it
+        killServer()
+        atexit.unregister(killServer)
 
 
     def __initConfigDir(self, cfgdir):
@@ -605,7 +669,7 @@ class NotebookServer(threading.Thread):
         cfgenv = {
             'fsleyes_nbserver_port'       : defaultPort,
             'fsleyes_nbserver_token'      : self.__token,
-            'fsleyes_nbserver_dir'        : op.expanduser('~'),
+            'fsleyes_nbserver_dir'        : os.getcwd(),
             'fsleyes_nbserver_static_dir' : cfgdir,
             'fsleyes_nbextension_dir'     : nbextdir,
             'fsleyes_kernel_connfile'     : self.__connfile,
@@ -629,3 +693,27 @@ class NotebookServer(threading.Thread):
         for fn, e in zip(files, envs):
             with open(fn, 'rt') as f: template = j2.Template(f.read())
             with open(fn, 'wt') as f: f.write(template.render(**e))
+
+
+def nbmain(argv):
+    """Wrapper around a Jupyter Notebook server entry point. Invoked by the
+    :class:`NotebookServer`, via a hook in :func:`fsleyes.main.main`.
+    """
+
+    if argv[:2] != ['notebook', 'server']:
+        raise RuntimeError('argv does not look like notebook main arguments '
+                           '(first args are not \'notebook server\')')
+
+    argv = argv[2:]
+
+    # run the notebook server
+    from notebook.notebookapp import main as nbmain
+
+    # first argument is a path
+    # to add to the PYTHONPATH.
+    # See NotebookServer.run.
+    sys.path.insert(0, argv[0])
+
+    # remaining arguments are passed
+    # through to notebookapp.main
+    return nbmain(argv=argv[1:])
