@@ -5,7 +5,7 @@
 # Author: Paul McCarthy <pauldmccarthy@gmail.com>
 #
 """This module provides the :class:`GLMesh` class, a :class:`.GLObject` used
-to render :class:`.TriangleMesh` overlays.
+to render :class:`.Mesh` overlays.
 """
 
 
@@ -13,19 +13,19 @@ import numpy        as np
 import numpy.linalg as npla
 import OpenGL.GL    as gl
 
-from . import                 globject
-import fsl.utils.transform as transform
-import fsleyes.gl          as fslgl
-import fsleyes.gl.trimesh  as trimesh
-import fsleyes.gl.routines as glroutines
-import fsleyes.gl.textures as textures
+from . import                  globject
+import fsl.data.utils       as dutils
+import fsl.transform.affine as affine
+import fsleyes.gl           as fslgl
+import fsleyes.gl.routines  as glroutines
+import fsleyes.gl.textures  as textures
 
 
 class GLMesh(globject.GLObject):
     """The ``GLMesh`` class is a :class:`.GLObject` which encapsulates the
     logic required to draw 2D slices, and 3D renderings, of a
-    :class:`.TriangleMesh` overlay.  The ``GLMesh`` class assumes that the
-    :class:`.Display` instance associated with the ``TriangleMesh`` overlay
+    :class:`.Mesh` overlay.  The ``GLMesh`` class assumes that the
+    :class:`.Display` instance associated with the ``Mesh`` overlay
     holds a reference to a :class:`.MeshOpts` instance, which contains
     ``GLMesh`` specific display settings.
 
@@ -60,12 +60,21 @@ class GLMesh(globject.GLObject):
     *Outlines*
 
 
-    If ``outline is True or vertexData is not None``, the
-    :func:`.trimesh.mesh_plane` function is used to calculate the intersection
-    of the mesh triangles with the viewing plane. Theese lines are then
+    If ``outline is True or vertexData is not None``, the intersection of the
+    mesh triangles with the viewing plane is calculated. These lines are then
     rendered as ``GL_LINES`` primitives.
 
 
+    When a mesh outline is drawn on a canvas, the calculated line vertices,
+    and corresponding mesh faces (triangles) are cached via the
+    :meth:`.OverlayList.setData` method. This is so that other parts of
+    FSLeyes can access this information if necessary (e.g. the
+    :class:`.OrthoViewProfile` provides mesh-specific interaction). For a
+    canvas which is showing a plane orthogonal to display axis X, Y or Z, this
+    data will be given a key of ``'crosssection_0'``, ``'crosssection_1'``, or
+    ``'crosssection_2'`` respectively.
+
+
     *Colouring*
 
 
@@ -107,8 +116,8 @@ class GLMesh(globject.GLObject):
 
      - ``preDraw(GLMesh)``: Prepare for drawing (e.g. load shaders)
 
-     - ``draw(GLMesh, glType, vertices, indices=None, normals=None,
-         vdata=None)``: Draws mesh using shaders.
+     - ``draw(GLMesh, glType, vertices, indices=None, normals=None, vdata=None)``:  # noqa
+        Draws mesh using shaders.
 
      - ``postDraw(GLMesh)``: Clean up after drawing
 
@@ -122,19 +131,32 @@ class GLMesh(globject.GLObject):
     data, shader programs are used which colour the mesh, and also apply a
     simple lighting effect.
 
+
+    **Flat shading**
+
+    When the :attr:`.MeshOpts.flatShading` property is active (only possible
+    when 3D rendering), the mesh vertices and indices need to be modified.
+    In order to achive flat shading, each vertex of a triangle must be given
+    the same vertex data value. This means that vertices cannot be shared by
+    different triangles. So when the ``flatShading`` property changes, the
+    :meth:`updateVertices` method will re-generate the vertices and indices
+    accordingly. The :meth:`getVertexData` method will modify the vertex
+    data that gets passed to the shader accordingly too.
     """
 
 
-    def __init__(self, overlay, displayCtx, canvas, threedee):
+    def __init__(self, overlay, overlayList, displayCtx, canvas, threedee):
         """Create a ``GLMesh``.
 
-        :arg overlay:    A :class:`.TriangleMesh` overlay.
-        :arg displayCtx: The :class:`.DisplayContext` managing the scene.
-        :arg canvas:     The canvas drawing this ``GLMesh``.
-        :arg threedee:   2D or 3D rendering.
+        :arg overlay:     A :class:`.Mesh` overlay.
+        :arg overlayList: The :class:`.OverlayList`
+        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
+        :arg canvas:      The canvas drawing this ``GLMesh``.
+        :arg threedee:    2D or 3D rendering.
         """
 
-        globject.GLObject.__init__(self, overlay, displayCtx, canvas, threedee)
+        globject.GLObject.__init__(
+            self, overlay, overlayList, displayCtx, canvas, threedee)
 
         self.flatShader   = None
         self.dataShader   = None
@@ -144,7 +166,8 @@ class GLMesh(globject.GLObject):
         # rendering model cross sections.
         # This texture is kept at display/
         # screen resolution
-        self.renderTexture = textures.RenderTexture(self.name, gl.GL_NEAREST)
+        self.renderTexture = textures.RenderTexture(
+            self.name, interp=gl.GL_NEAREST)
 
         # Mesh overlays are coloured:
         #
@@ -171,7 +194,7 @@ class GLMesh(globject.GLObject):
 
     def destroy(self):
         """Must be called when this ``GLMesh`` is no longer needed. Removes
-        some property listeners and destroys the colour map texturtes and
+        some property listeners and destroys the colour map textures and
         off-screen :class:`.RenderTexture`.
         """
 
@@ -237,14 +260,17 @@ class GLMesh(globject.GLObject):
         def refresh(*a):
             self.notify()
 
+        self.overlay.register(name, vertices, 'vertices')
         opts   .addListener('bounds',           name, vertices,    weak=False)
         opts   .addListener('colour',           name, shader,      weak=False)
         opts   .addListener('outline',          name, refresh,     weak=False)
         opts   .addListener('outlineWidth',     name, refresh,     weak=False)
         opts   .addListener('wireframe',        name, refresh,     weak=False)
         opts   .addListener('vertexData',       name, shader,      weak=False)
+        opts   .addListener('modulateData',     name, shader,      weak=False)
         opts   .addListener('vertexDataIndex',  name, shader,      weak=False)
         opts   .addListener('clippingRange',    name, shader,      weak=False)
+        opts   .addListener('modulateRange',    name, shader,      weak=False)
         opts   .addListener('invertClipping',   name, shader,      weak=False)
         opts   .addListener('discardClipped',   name, shader,      weak=False)
         opts   .addListener('cmap',             name, refreshCmap, weak=False)
@@ -253,9 +279,12 @@ class GLMesh(globject.GLObject):
         opts   .addListener('cmapResolution',   name, refreshCmap, weak=False)
         opts   .addListener('interpolateCmaps', name, refreshCmap, weak=False)
         opts   .addListener('invert',           name, refreshCmap, weak=False)
+        opts   .addListener('gamma',            name, refreshCmap, weak=False)
         opts   .addListener('displayRange',     name, refreshCmap, weak=False)
         opts   .addListener('useLut',           name, shader,      weak=False)
         opts   .addListener('lut',              name, registerLut, weak=False)
+        opts   .addListener('modulateAlpha',    name, shader,      weak=False)
+        opts   .addListener('flatShading',      name, vertices,    weak=False)
         display.addListener('alpha',            name, refreshCmap, weak=False)
 
         if self.threedee:
@@ -271,14 +300,17 @@ class GLMesh(globject.GLObject):
         """Called by :meth:`destroy`. Removes all of the listeners added by
         the :meth:`addListeners` method.
         """
+        self.overlay.deregister(self.name, 'vertices')
         self.opts   .removeListener('bounds',           self.name)
         self.opts   .removeListener('colour',           self.name)
         self.opts   .removeListener('outline',          self.name)
         self.opts   .removeListener('outlineWidth',     self.name)
         self.opts   .removeListener('wireframe',        self.name)
         self.opts   .removeListener('vertexData',       self.name)
+        self.opts   .removeListener('modulateData',     self.name)
         self.opts   .removeListener('vertexDataIndex',  self.name)
         self.opts   .removeListener('clippingRange',    self.name)
+        self.opts   .removeListener('modulateRange',    self.name)
         self.opts   .removeListener('invertClipping',   self.name)
         self.opts   .removeListener('discardClipped',   self.name)
         self.opts   .removeListener('cmap',             self.name)
@@ -287,9 +319,12 @@ class GLMesh(globject.GLObject):
         self.opts   .removeListener('cmapResolution',   self.name)
         self.opts   .removeListener('interpolateCmaps', self.name)
         self.opts   .removeListener('invert',           self.name)
+        self.opts   .removeListener('gamma',            self.name)
         self.opts   .removeListener('displayRange',     self.name)
         self.opts   .removeListener('useLut',           self.name)
         self.opts   .removeListener('lut',              self.name)
+        self.opts   .removeListener('modulateAlpha',    self.name)
+        self.opts   .removeListener('flatShading',      self.name)
         self.display.removeListener('alpha',            self.name)
 
         if self.threedee:
@@ -329,20 +364,41 @@ class GLMesh(globject.GLObject):
         """
 
         overlay  = self.overlay
+        opts     = self.opts
+        threedee = self.threedee
         vertices = overlay.vertices
         indices  = overlay.indices
         normals  = self.overlay.vnormals
-        xform    = self.opts.getCoordSpaceTransform()
+        vdata    = opts.getVertexData('vertex')
+        xform    = opts.getTransform('mesh', 'display')
 
         if not np.all(np.isclose(xform, np.eye(4))):
-            vertices = transform.transform(vertices, xform)
+            vertices = affine.transform(vertices, xform)
 
             if self.threedee:
-                nmat    = transform.invert(xform).T
-                normals = transform.transform(normals, nmat, vector=True)
+                nmat    = affine.invert(xform).T
+                normals = affine.transform(normals, nmat, vector=True)
 
-        self.vertices = np.array(vertices,          dtype=np.float32)
-        self.indices  = np.array(indices.flatten(), dtype=np.uint32)
+        self.origIndices = indices
+        indices          = np.asarray(indices.flatten(), dtype=np.uint32)
+
+        # If flatShading is active, we cannot share
+        # vertices between triangles, so we generate
+        # a set of unique vertices for each triangle,
+        # and then re-generate the triangle indices.
+        # The original indices are saved above, as
+        # they will be used by the getVertexData
+        # method to duplicate the vertex data.
+        if threedee and (vdata is not None) and opts.flatShading:
+            self.vertices = vertices[indices].astype(np.float32)
+            self.indices  = np.arange(0, len(self.vertices), dtype=np.uint32)
+            normals       = normals[indices, :]
+        else:
+            self.vertices = np.asarray(vertices, dtype=np.float32)
+            self.indices  = indices
+
+        self.vertices = dutils.makeWriteable(self.vertices)
+        self.indices  = dutils.makeWriteable(self.indices)
 
         if self.threedee:
             self.normals = np.array(normals, dtype=np.float32)
@@ -363,7 +419,7 @@ class GLMesh(globject.GLObject):
         # transform, thus we are assuming that
         # the MVP matrix does not have any
         # negative scales.
-        xform = self.opts.getCoordSpaceTransform()
+        xform = self.opts.getTransform('mesh', 'display')
 
         if npla.det(xform) > 0: return gl.GL_CCW
         else:                   return gl.GL_CW
@@ -377,6 +433,27 @@ class GLMesh(globject.GLObject):
                 self.opts.bounds.getHi())
 
 
+    def draw2DOutlineEnabled(self):
+        """Only relevent for 2D rendering. Returns ``True`` if outline mode
+        should be used, ``False`` otherwise.
+        """
+
+        opts    = self.opts
+        overlay = self.overlay
+
+        return ((overlay.trimesh is not None) and
+                (opts.outline or opts.vertexData is not None))
+
+
+    def needShader(self):
+        """Returns ``True`` if a shader should be loaded, ``False`` otherwise.
+        Relevant for both 2D and 3D rendering.
+        """
+        return (self.threedee or
+                (self.draw2DOutlineEnabled() and
+                 self.opts.vertexData is not None))
+
+
     def preDraw(self, xform=None, bbox=None):
         """Overrides :meth:`.GLObject.preDraw`. Performs some pre-drawing
         configuration, which might involve loading shaders, and/or setting the
@@ -384,21 +461,24 @@ class GLMesh(globject.GLObject):
         current viewport size.
         """
 
-        opts       = self.opts
-        flat       = self.opts.vertexData is None
-        useShader  = self.threedee or (opts.vertexData is not None)
-        useTexture = not (self.threedee or opts.outline or useShader)
+        useShader  = self.needShader()
+        outline    = self.draw2DOutlineEnabled()
+        useTexture = not (self.threedee or outline)
 
+        # A shader program is used either in 3D, or
+        # in 2D when some vertex data is being shown
         if useShader:
             fslgl.glmesh_funcs.preDraw(self)
 
-            if not flat:
+            if self.opts.vertexData is not None:
                 if self.opts.useLut:
                     self.lutTexture.bindTexture(gl.GL_TEXTURE0)
                 else:
                     self.cmapTexture   .bindTexture(gl.GL_TEXTURE0)
                     self.negCmapTexture.bindTexture(gl.GL_TEXTURE1)
 
+        # An off-screen texture is used when
+        # drawing off-screen textures in 2D
         if useTexture:
             size   = gl.glGetIntegerv(gl.GL_VIEWPORT)
             width  = size[2]
@@ -406,24 +486,19 @@ class GLMesh(globject.GLObject):
 
             # We only need to resize the texture when
             # the viewport size/quality changes.
-            if self.renderTexture.getSize() != (width, height):
-                self.renderTexture.setSize(width, height)
+            if self.renderTexture.shape != (width, height):
+                self.renderTexture.shape = width, height
 
 
     def draw2D(self, zpos, axes, xform=None, bbox=None):
         """Overrids :meth:`.GLObject.draw2D`. Draws a 2D slice of the
-        :class:`.TriangleMesh`, at the specified Z location.
+        :class:`.Mesh`, at the specified Z location.
         """
 
-        opts          = self.opts
+        overlay       = self.overlay
         lo,  hi       = self.getDisplayBounds()
         xax, yax, zax = axes
-
-        # If vertexData is set, we ignore
-        # the outline setting - meshes can
-        # only be coloured by vertexData
-        # in outline mode
-        outline = opts.outline or opts.vertexData is not None
+        outline       = self.draw2DOutlineEnabled()
 
         # Mesh is 2D, and is
         # perpendicular to
@@ -438,6 +513,12 @@ class GLMesh(globject.GLObject):
         if not is2D and (zpos < lo[zax] or zpos > hi[zax]):
             return
 
+        # the calculateIntersection method caches
+        # cross section vertices - make sure they're
+        # cleared in case we are not doing an outline
+        # draw.
+        self.overlayList.setData(overlay, 'crosssection_{}'.format(zax), None)
+
         if is2D:
             self.draw2DMesh(xform, bbox)
 
@@ -466,7 +547,11 @@ class GLMesh(globject.GLObject):
         idxs      = self.indices
         normals   = self.normals
         blo, bhi  = self.getDisplayBounds()
-        vdata     = opts.getVertexData()
+        vdata     = self.getVertexData('vertex')
+        mdata     = self.getVertexData('modulate')
+
+        if mdata is None:
+            mdata = vdata
 
         is2D = np.isclose(bhi[2], blo[2])
 
@@ -481,8 +566,10 @@ class GLMesh(globject.GLObject):
             gl.glPushMatrix()
             gl.glMultMatrixf(np.array(xform, dtype=np.float32).ravel('F'))
 
-        if is2D: enable = (gl.GL_DEPTH_TEST)
-        else:    enable = (gl.GL_DEPTH_TEST, gl.GL_CULL_FACE)
+        if is2D or opts.wireframe:
+            enable = (gl.GL_DEPTH_TEST)
+        else:
+            enable = (gl.GL_DEPTH_TEST, gl.GL_CULL_FACE)
 
         gl.glDisable(gl.GL_CULL_FACE)
         with glroutines.enabled(enable):
@@ -495,7 +582,8 @@ class GLMesh(globject.GLObject):
                 verts,
                 normals=normals,
                 indices=idxs,
-                vdata=vdata)
+                vdata=vdata,
+                mdata=mdata)
 
         if xform is not None:
             gl.glPopMatrix()
@@ -507,10 +595,7 @@ class GLMesh(globject.GLObject):
         :func:`.gl21.glmesh_funcs.postDraw` function.
         """
 
-        opts      = self.opts
-        useShader = self.threedee or (opts.vertexData is not None)
-
-        if useShader:
+        if self.needShader():
             fslgl.glmesh_funcs.postDraw(self)
 
             if self.opts.vertexData is not None:
@@ -543,13 +628,17 @@ class GLMesh(globject.GLObject):
             zpos, axes, bbox)
 
         if vertXform is not None:
-            xform = transform.concat(xform, vertXform)
+            xform = affine.concat(xform, vertXform)
 
-        vdata     = self.getVertexData(faces, dists)
+        vdata     = self.getVertexData('vertex',   faces, dists)
+        mdata     = self.getVertexData('modulate', faces, dists)
         useShader = vdata is not None
         vertices  = vertices.reshape(-1, 3)
         nvertices = vertices.shape[0]
 
+        if mdata is None:
+            mdata = vdata
+
         gl.glMatrixMode(gl.GL_MODELVIEW)
         gl.glPushMatrix()
         gl.glMultMatrixf(np.array(xform, dtype=np.float32).ravel('F'))
@@ -571,23 +660,28 @@ class GLMesh(globject.GLObject):
                 self,
                 gl.GL_LINES,
                 vertices,
-                vdata=vdata)
+                vdata=vdata,
+                mdata=mdata)
 
         gl.glPopMatrix()
 
 
     def draw2DMesh(self, xform=None, bbox=None):
         """Not to be confused with :meth:`draw2D`.  Called by :meth:`draw2D`
-        for :class:`.TriangleMesh` overlays which are actually 2D (with a flat
+        for :class:`.Mesh` overlays which are actually 2D (with a flat
         third dimension).
         """
 
         opts      = self.opts
-        vdata     = opts.getVertexData()
-        useShader = vdata is not None
+        vdata     = opts.getVertexData('vertex')
+        mdata     = opts.getVertexData('modulate')
+        useShader = self.needShader()
         vertices  = self.vertices
         faces     = self.indices
 
+        if mdata is None:
+            mdata = vdata
+
         if opts.outline:
             gl.glLineWidth(opts.outlineWidth)
             gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
@@ -610,13 +704,16 @@ class GLMesh(globject.GLObject):
 
         # Coloured from vertex data
         else:
+            # TODO separate modulateDataIndex?
             vdata = vdata[:, opts.vertexDataIndex]
+            mdata = mdata[:, 0]
             fslgl.glmesh_funcs.draw(
                 self,
                 gl.GL_TRIANGLES,
                 vertices,
                 indices=faces,
-                vdata=vdata)
+                vdata=vdata,
+                mdata=mdata)
 
 
     def drawCrossSection(self, zpos, axes, lo, hi, dest):
@@ -692,7 +789,7 @@ class GLMesh(globject.GLObject):
         # has a negative determinant, it means
         # the back faces will be facing the camera,
         # so we need to render the back faces first.
-        if npla.det(opts.getCoordSpaceTransform()) > 0:
+        if npla.det(opts.getTransform('mesh', 'display')) > 0:
             faceOrder = [gl.GL_FRONT, gl.GL_BACK]
         else:
             faceOrder = [gl.GL_BACK,  gl.GL_FRONT]
@@ -784,9 +881,9 @@ class GLMesh(globject.GLObject):
 
 
     def calculateIntersection(self, zpos, axes, bbox=None):
-        """Uses the :func:`.trimesh.mesh_plane` and
-        :func:`.trimesh.points_to_barycentric` functions to calculate the
-        intersection of the mesh with the viewing plane at the given ``zpos``.
+        """Uses the :func:`.Mesh.planeIntersection` method to
+        calculate the intersection of the mesh with the viewing plane at
+        the given ``zpos``.
 
         :arg zpos:  Z axis coordinate at which the intersection is to be
                     calculated
@@ -802,7 +899,7 @@ class GLMesh(globject.GLObject):
                      a line for every intersected face (triangle) in the mesh.
 
                    - A ``(n, 3)`` array containing the intersected faces
-                     (indices into the :attr:`.TriangleMesh.vertices` array).
+                     (indices into the :attr:`.Mesh.vertices` array).
 
                    - A ``(n, 2, 3)`` array containing the barycentric
                      coordinates of the intersection line vertices.
@@ -811,6 +908,12 @@ class GLMesh(globject.GLObject):
                      for transforming the line vertices into the display
                      coordinate system. May be ``None``, indicating that
                      no transformation is necessary.
+
+        .. note:: The line vertices, and corresponding mesh triangle face
+                  indices, which make up the cross section are saved
+                  via the :meth:`.OverlayList.setData` method, with a key
+                  ``'crosssection_[zax]'``, where ``[zax]`` is set to the
+                  index of the display Z axis.
         """
 
         overlay     = self.overlay
@@ -821,69 +924,76 @@ class GLMesh(globject.GLObject):
         origin[zax] = zpos
         normal[zax] = 1
 
-        if opts.refImage is not None:
+        vertXform = opts.getTransform(           'mesh',    'display')
+        origin    = opts.transformCoords(origin, 'display', 'mesh')
+        normal    = opts.transformCoords(normal, 'display', 'mesh',
+                                         vector=True)
 
-            ropts  = opts.displayCtx.getOpts(opts.refImage)
-            origin = ropts.transformCoords(origin,
-                                           ropts.transform,
-                                           opts.coordSpace)
+        # TODO use bbox to constrain? This
+        #      would be nice, but is not
+        #      supported by trimesh.
+        lines, faces, dists = overlay.planeIntersection(
+            normal, origin, distances=True)
 
-            vertXform = ropts.getTransform(opts.coordSpace, ropts.transform)
+        lines = np.asarray(lines, dtype=np.float32)
+        faces = np.asarray(faces, dtype=np.uint32)
 
-        else:
-            vertXform = None
-
-        # TODO use bbox to constrain?
-        lines, faces = trimesh.mesh_plane(
-            overlay.vertices,
-            overlay.indices,
-            plane_normal=normal,
-            plane_origin=origin,
-            return_faces=True)
+        # cache the line vertices for other
+        # things which might be interested.
+        # See, for example, OrthoViewProfile
+        # pick mode methods.
+        self.overlayList.setData(overlay,
+                                 'crosssection_{}'.format(zax),
+                                 (lines, faces))
 
         faces = overlay.indices[faces]
 
-        # Calculate the barycentric coordinates
-        # (distance from triangle vertices) for
-        # each intersection line - this is used
-        # for interpolating vertex colours.
-
-        triangles = overlay.vertices[faces].repeat(2, axis=0)
-        points    = lines.reshape((-1, 3))
-
-        if triangles.size > 0:
-            dists = trimesh.points_to_barycentric(triangles, points)
-            dists = dists.reshape((-1, 2, 3))
-        else:
-            dists = np.zeros((0, 2, 3))
-
         return lines, faces, dists, vertXform
 
 
-    def getVertexData(self, faces, dists):
-        """If :attr:`.MeshOpts.vertexData` is not ``None``, this method
-        returns the vertex data to use for the line segments calculated
-        in the :meth:`calculateIntersection` method.
+    def getVertexData(self, vdtype, faces=None, dists=None):
+        """If :attr:`.MeshOpts.vertexData` (or :attr:`.MeshOpts.modulateData`)
+        is not ``None``, this method returns the vertex data to use for the
+        current vertex data index.
 
-        The ``dists`` array contains barycentric coordinates for each line
-        vertex, and is used to linearly interpolate between the values of the
-        vertices of the intersected triangles (defined in ``faces``).
+        ``faces`` and ``dists`` are used by the :meth:`drawOutline` method,
+        which draws a cross-section of the mesh. The ``dists`` array contains
+        barycentric coordinates for each line vertex, and is used to linearly
+        interpolate between the values of the vertices of the intersected
+        triangles (defined in ``faces``).
 
         If ``MeshOpts.vertexData is None``, this method returns ``None``.
+
+        :arg vdtype: Use ``'vertex'`` for :attr:`.MeshOpts.vertexData`, or
+                     ``'modulate'`` for :attr:`.MeshOpts.modulateData`.
         """
 
         opts  = self.opts
-        vdata = opts.getVertexData()
+        vdata = opts.getVertexData(vdtype)
 
         if vdata is None:
             return None
 
-        vdata = vdata[:, opts.vertexDataIndex]
+        # TODO separate modulateDataIndex for modulateData?
+        if vdtype == 'vertex': vdata = vdata[:, opts.vertexDataIndex]
+        else:                  vdata = vdata[:, 0]
 
-        vdata = vdata[faces].repeat(2, axis=0).reshape(-1, 2, 3)
-        vdata = (vdata * dists).reshape(-1, 3).sum(axis=1)
+        # when flat-shading, we colour each
+        # triangle according to the data
+        # value for its first vertex. This
+        # should really be pre-generated
+        # whenever the vertex data or flat-
+        # shading option is changed, but it
+        # is quick for typical surfaces, so
+        # I'm not bothering for now.
+        if self.threedee and (vdata is not None) and opts.flatShading:
+            vdata = vdata[self.origIndices[:, 0]].repeat(3)
 
-        return vdata
+        if faces is not None and dists is not None:
+            vdata = vdata[faces].repeat(2, axis=0).reshape(-1, 2, 3)
+            vdata = (vdata * dists).reshape(-1, 3).sum(axis=1)
+
+        return np.asarray(vdata, np.float32)
 
 
     def refreshCmapTextures(self, *a, **kwa):
@@ -908,6 +1018,7 @@ class GLMesh(globject.GLObject):
         interp  = opts.interpolateCmaps
         res     = opts.cmapResolution
         negCmap = opts.negativeCmap
+        gamma   = opts.realGamma(opts.gamma)
         invert  = opts.invert
         dmin    = opts.displayRange[0]
         dmax    = opts.displayRange[1]
@@ -919,6 +1030,7 @@ class GLMesh(globject.GLObject):
                              invert=invert,
                              alpha=alpha,
                              resolution=res,
+                             gamma=gamma,
                              interp=interp,
                              displayRange=(dmin, dmax))
 
@@ -926,6 +1038,7 @@ class GLMesh(globject.GLObject):
                                 invert=invert,
                                 alpha=alpha,
                                 resolution=res,
+                                gamma=gamma,
                                 interp=interp,
                                 displayRange=(dmin, dmax))
 
@@ -969,13 +1082,28 @@ class GLMesh(globject.GLObject):
 
         if dopts.useLut:
             delta     = 1.0 / (dopts.lut.max() + 1)
-            cmapXform = transform.scaleOffsetXform(delta, 0.5 * delta)
+            cmapXform = affine.scaleOffsetXform(delta, 0.5 * delta)
         else:
             cmapXform = self.cmapTexture.getCoordinateTransform()
 
+        # calculate a scale+offset which transforms
+        # modulate alpha value from the data range
+        # into an alpha value, according to the
+        # modulateRange
+        modlo, modhi = dopts.modulateRange
+        modRange     = modhi - modlo
+        if modRange == 0:
+            modScale  = 1
+            modOffset = 0
+        else:
+            modScale  = 1 / modRange
+            modOffset = -modlo / modRange
+
         fslgl.glmesh_funcs.updateShaderState(
             self,
             useNegCmap=useNegCmap,
             cmapXform=cmapXform,
+            modScale=modScale,
+            modOffset=modOffset,
             flatColour=flatColour,
             lightPos=lightPos)
